signature PRINT =
sig
  (* print an expression *)
  val expToString : Syntax.exp -> string
  val typToString : Syntax.typ -> string
  val stateToString : Syntax.exp -> string
end

structure Print :> PRINT =
struct

  open Syntax

  (* NOTE: you can change this to suppress the type annotations on terms *)
  val showTypesInTerms = true

  (* NOTE: you can change this to determine whether bool/nat/list values
           get printed out expanded into products/sums/recursive types or not.

           note that patterns are always expanded because they don't
           have enough type information to un-expand
  *)
  val expandValues = false

  fun typ t = 
      case t of 
          Arrow(t1,t2) => "( " ^ typ t1 ^ " -> " ^ typ t2 ^ " )"
        | Nat => "nat"
        | List => "list"

  fun maybetyp t = if showTypesInTerms then typ t else ""

  fun exp0 (tab, e) = case e of
      App(e1, e2) =>
          exp0 (tab, e1) ^ "@" ^ expinf (tab, e2)
    | _ =>
          let
                            
                fun bin (e : exp) : int option =
                    case e of
                        Zero =>  SOME 0 
                      | Succ e => Option.map (fn x => 1 + x) (bin e) 
                      | _ => NONE
                            
                fun bil (e : exp) : (int list) option =
                    case e of
                        Nil => SOME [] 
                      | Cons(e1,e2) =>
                            (case (bin e1, bil e2) of
                                 (SOME x, SOME xs) => SOME (x::xs)
                               | _ => NONE)
                      | _ => NONE
                                
          in 
              if expandValues
              then expinf (tab, e)
              else (* detect nats and lists and uncompile them *)
                  (case bin e of
                       SOME v => Int.toString v
                     | NONE => (case bil e of
                                    SOME v => "[" ^ List.foldr (fn (x,y) => Int.toString x ^ "," ^ y) "]" v
                                  | NONE => expinf (tab, e)))
          end
                       
  and expinf (tab, e) = case e of
      Var(x) => nameToString x
    | Lam(t,(x,e)) => "(lam " ^ nameToString x ^ " : " ^ maybetyp t ^ "." ^ exp0 (tab,e) ^")"
    | Let(e1, t, (x, e2)) =>
          let val tab' = tab ^ "    "
          in "let " ^ nameToString x ^ " : " ^ maybetyp t ^ " be " ^ exp0 (tab, e1) ^ " in" ^
              tab' ^ exp0 (tab', e2) ^
              tab ^ "end" end
    | Zero => "zero"
    | Succ(e) => "succ(" ^ exp0 (tab,e) ^ ")"
    | Nil => "zero"
    | Cons(e1, e2) => "cons< " ^ exp0 (tab,e1) ^ " , " ^ exp0 (tab,e2) ^ ">"
    | NatCase (e, e0, (x,e1)) =>
              "natcase " ^ exp0 (tab,e) 
              ^ " { z => " ^ exp0 (tab,e0) ^ " | s " 
              ^ nameToString x ^ " => " ^ exp0 (tab,e1) ^ "}"
    | ListCase (e, e0, (x,y,e1)) =>
              "listcase " ^ exp0 (tab,e) 
              ^ " { nil => " ^ exp0 (tab,e0) ^ " | cons< " 
              ^ nameToString x ^ " , " ^ nameToString y ^ "> => " ^ exp0 (tab,e1) ^ "}"
    | Fix(t,(x,e)) => "(fix " ^ nameToString x ^ " : " ^ maybetyp t ^ "." ^ exp0 (tab,e) ^")"
    | Subst(e, s) => exp0(tab, e) ^ "[ " ^ List.foldr (fn ((e',x'),r) => exp0(tab, e') ^ " / " ^ nameToString x' ^ "," ^ r) "" s ^ "]"
    | _ => "(" ^ exp0 (tab, e) ^ ")"

  fun expToString t = exp0 ("\n", t)
  fun typToString t = typ t

  fun stateToString (e) = "Expression: " ^ expToString(e)

      
end;  (* functor PrintFun *)
